/* $Id: async_lookup.c,v 1.5 1997/07/11 23:38:39 gjb Exp $
 * async_lookup.c
 * (C) 1997 Greg J. Badros -- All Rights Reserved
 * <gjb@cs.washington.edu> 
 * Seattle, WA  USA
 * http://www.cs.washington.edu/homes/gjb
 */

#define NFS_NEED_XDR_TYPES
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/fcntl.h>
#include <linux/stat.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/pagemap.h>
#include <linux/sunrpc/clnt.h>
#include <linux/nfs_fs.h>
/*#include <linux/errno.h> */

#include <asm/segment.h>
#include <asm/system.h>

#include "profile.h"
#include "params.h"
#include "nfs_cache.h"
#include "cache_failed_lookups.h"

struct nfs_lreq {
  struct inode *		la_inode;	/* inode from which to read */
  struct nfs_diropargs	la_args; 	        /* XDR argument struct */
  struct nfs_diropok	la_res;		/* ... and result struct */
	kdev_t dev;
	unsigned long ino;
	struct nfs_fh la_fh; /* the filehandle result */
	struct nfs_fattr la_fattr; /* the attributes result */
};


/* #define NFS_LOOKUP_RESULT_DEBUG */

static void
nfs_lookup_result(struct rpc_task *task)
{
	struct nfs_lreq	*req = (struct nfs_lreq *) task->tk_calldata;
	const char *name = req->la_args.name;
	int result = task->tk_status;
#ifdef NFS_LOOKUP_RESULT_DEBUG
	struct nfs_fh *fh = &req->la_fh;
	struct nfs_fattr *fattr = &req->la_fattr;
#endif
	nfs_failed_lookup_result(req->dev,req->ino,name,result);
#ifdef NFS_LOOKUP_RESULT_DEBUG
	printk("NFS: %s received callback for %s @%ld, (%d,%ld), fh = (%d,%d,%d,%d,%d), result %d: size %d\n",
				 __FUNCTION__,name,jiffies,
				 req->dev,req->ino,
				 fh->data[0], fh->data[1], fh->data[2], fh->data[3], fh->data[4],
				 result,fattr->size);
#endif
	rpc_release_task(task);
	kfree(req);
}


/* nfs_proc_lookup also had "struct super_block *sb" as an argument,
   but it doesn't appear to be used */
int
nfs_async_lookup(struct inode *inodeDir, const char *name)
{
	int			status;
	struct nfs_lreq *req;

	alnfs_printk("NFS async lookup %s, @ %ld\n", name, jiffies);

	if (!(req = (struct nfs_lreq *) rpc_allocate(RPC_TASK_ASYNC,sizeof(*req)))) {
		printk("NFS: failed rpc_allocate for async lookup\n");
		return -1;
	}

	req->la_args.fh = NFS_FH(inodeDir);
	req->la_args.name = name;
	req->dev = inodeDir->i_dev;
	req->ino = inodeDir->i_ino;

	req->la_res.fh = &req->la_fh;
	req->la_res.fattr = &req->la_fattr;

	/* FIXGJB: still have the dir vs. inode_number of the dir problem */
	status = rpc_do_call(NFS_CLIENT(inodeDir), NFSPROC_LOOKUP,
											 &req->la_args, &req->la_res, RPC_TASK_ASYNC,
											 nfs_lookup_result, req);
	if (status > 0)
		alnfs_printk("NFS: %s failed to enqueue async lookup request\n",
								 __FUNCTION__);
	
	return status;
}	
