/* $Id: cache_failed_lookups.h,v 1.5 1997/07/11 23:40:30 gjb Exp $
 * cache_failed_lookups.h 
 * Greg Badros 
 * <gjb@cs.washington.edu> 
 * Seattle, WA  USA
 * http://www.cs.washington.edu/homes/gjb
 */

#ifndef _CACHE_FAILED_LOOKUPS_H_
#define _CACHE_FAILED_LOOKUPS_H_
#include "profile.h"
#include "params.h"

#define NO_CREATE 2
#define NEED_NOT_FIND 1
#define MUST_FIND 0

#define NFS_FL_MAX_NAME_LEN 20
#define NFS_FL_WINDOW_SIZE 32
#define NFS_PRED_WINDOW_SIZE 10
#define NFS_FL_MAX_TOPLEVEL_DIRS NFS_FL_WINDOW_SIZE
#define NFS_FL_MAX_TRANSITION_DIRS 8
#define NFS_FL_LOOKAHEAD 4
#define NFS_FL_MAX_PREDICTIONS 2
/* only make prediction when we're k% sure */
#define NFS_P_REQUIRED_TO_PREDICT 20

#ifdef DEBUG_NFSFLC_PRINTK
#define nfsflc_printk(s) printk("NFS_FLC: %s (%d,%ld),%s; " s "\n",__FUNCTION__,dev,ino,name)
#else
#define nfsflc_printk(s)
#endif

/* nfs_handle_failed_lookup:
   Updates prediction data structures, and makes asynchronous lookup calls
   */
extern void nfs_handle_failed_lookup(struct inode *dir,char *name,int error);

/* nfs_failed_lookup_cache_lookup:
   Return -ENOENT if it's in our cache of failed lookups,
   or 0 if not */
extern int nfs_failed_lookup_cache_lookup(struct inode *dir, char *name);

extern int nfs_failed_lookup_result(kdev_t dev, unsigned long ino, const char *name, int status);

typedef struct nfs_dirid {
  kdev_t dev;
  unsigned long ino;
  struct inode *inode;
} nfs_dirid;
  
typedef struct nfs_failed_lookup {
  nfs_dirid dirid;
  char name[NFS_FL_MAX_NAME_LEN];
  unsigned long jiffies;
} nfs_failed_lookup;

typedef struct nfs_pred_lookup {
  nfs_dirid dirid;
  char name[NFS_FL_MAX_NAME_LEN];
  unsigned long jiffies;
  int fUsed;
  int fValid;
} nfs_pred_lookup;


typedef struct nfs_transition_nextdir {
  nfs_dirid dirid;
  unsigned long c;
} nfs_transition_nextdir;

typedef struct nfs_transitions {
  nfs_dirid dirid;
  /* c == number of times this directory is in the window */
  unsigned long c; 
  /* cTotTransitions == sum of the count fields of the target directories in rg */
  unsigned long cTotTransitions;
  nfs_transition_nextdir rg[NFS_FL_MAX_TRANSITION_DIRS];
} nfs_transitions;

#define QUEUE_DEC(x,s) (((x)+(s)+1-1)%((s)+1))
#define QUEUE_INC(x,s) (((x)+1)%((s)+1))
#define QUEUE_SIZE(hd,tl,s) (((tl) + (s) - (hd)) % (s))
#define QUEUE_FULL(hd,tl,s) ((hd) == QUEUE_INC((tl),(s)))

extern int cPredMadeTotal;
extern int cPredUnused;
extern int cPredCorrect;
extern int cPredHits;
extern int cRecentUseHit;
extern int cTotLookups;
extern int cPredOverflows;

#endif /* _CACHE_FAILED_LOOKUPS_H_ */
