/* $Id: failed-lookup-caching.c,v 1.13 1997/07/04 20:49:05 gjb Exp $
 * Greg Badros 
 * <gjb@cs.washington.edu> 
 * Seattle, WA  USA  {FIX: use env_vars}
 * http://www.cs.washington.edu/homes/gjb
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#ifndef lint
static char *rcsid = "$Header: /homes/gws/gjb/AFS/.CVS-551/linux/fs/nfs/lookup-data/failed-lookup-caching.c,v 1.13 1997/07/04 20:49:05 gjb Exp $";
#endif

#define NO_CREATE 2
#define NEED_NOT_FIND 1
#define MUST_FIND 0

#define TRUE (0==0)
#define FALSE (0==1)

#define HZ 60
#define NFS_FL_MAX_NAME_LEN 20
#define NFS_FL_WINDOW_SIZE 32
#define NFS_PRED_WINDOW_SIZE 10
#define NFS_FL_MAX_TOPLEVEL_DIRS NFS_FL_WINDOW_SIZE
#define NFS_FL_MAX_TRANSITION_DIRS 8
#define NFS_FL_LOOKAHEAD 4
#define NFS_FL_MAX_PREDICTIONS 2
/* only make prediction when we're k% sure */
#define NFS_P_REQUIRED_TO_PREDICT 20

typedef long kdev_t;

unsigned long nfs_fl_timeout = .20 * HZ;

struct inode {
  int foo;
};

typedef struct nfs_dirid {
  kdev_t dev;
  unsigned long ino;
  struct inode *inode;
} nfs_dirid;
  
typedef struct nfs_failed_lookup {
  nfs_dirid dirid;
  char name[NFS_FL_MAX_NAME_LEN];
  unsigned long jiffies;
} nfs_failed_lookup;

typedef struct nfs_pred_lookup {
  nfs_dirid dirid;
  char name[NFS_FL_MAX_NAME_LEN];
  unsigned long jiffies;
  int fUsed;
} nfs_pred_lookup;


typedef struct nfs_transition_nextdir {
  nfs_dirid dirid;
  unsigned long c;
} nfs_transition_nextdir;

typedef struct nfs_transitions {
  nfs_dirid dirid;
  /* c == number of times this directory is in the window */
  unsigned long c; 
  /* cTotTransitions == sum of the count fields of the target directories in rg */
  unsigned long cTotTransitions;
  nfs_transition_nextdir rg[NFS_FL_MAX_TRANSITION_DIRS];
} nfs_transitions;

#define QUEUE_DEC(x,s) (((x)+(s)+1-1)%((s)+1))
#define QUEUE_INC(x,s) (((x)+1)%((s)+1))
#define QUEUE_SIZE(hd,tl,s) (((tl) + (s) - (hd)) % (s))
#define QUEUE_FULL(hd,tl,s) ((hd) == QUEUE_INC((tl),(s)))

/* a queue window of the last failed lookups, uses cyclic-array impl.,
   lookups added to the tail, removed from the head */
static nfs_failed_lookup rgNFS_failures[NFS_FL_WINDOW_SIZE+1];
static int iflc_hd = 0;
static int iflc_tl = 0;

/* list of the directories and their transitions for the lookups in the
   above rgNFS_failures queue window.  These are in decreasing
   order of frequency (we do linear lookup, so we want dirs occurring
   most often to be cheapest to find */
static nfs_transitions rgNFS_transitions[NFS_FL_MAX_TOPLEVEL_DIRS+1];
static int cNFS_transitions = 0; /* # elem in above list */

static nfs_pred_lookup rgNFS_predictions[NFS_PRED_WINDOW_SIZE+1];
static int ip_hd = 0;
static int ip_tl = 0;


static int cPredMadeTotal = 0;
static int cPredUnused = 0;
static int cPredCorrect = 0;
static int cPredHits = 0;
static int cRecentUseHit = 0;
static int cTotLookups = 0;
static int cPredOverflows = 0;

static void nfs_fl_dump();
static void nfs_fl_dump_preds();

static int /* index into rgNFS_transitions[it].rg[] */
nfs_flt_target_find( int it, nfs_dirid dirid, int fHaveToFind )
{
  int i = 0;
  while (rgNFS_transitions[it].rg[i].dirid.dev != 0 &&
	 i < NFS_FL_MAX_TRANSITION_DIRS)
    {
    if (rgNFS_transitions[it].rg[i].dirid.dev == dirid.dev &&
	rgNFS_transitions[it].rg[i].dirid.ino == dirid.ino)
      return i;
    i++;
    }
  /* fHaveToFind is just there for debugging */
  assert(fHaveToFind);

  if (i == NFS_FL_MAX_TRANSITION_DIRS)
    { /* it's full, so cycle the spots with 1 counts, if any */
    /* FIXGJB */
    fprintf(stderr,"Filled... ignoring\n");
    return;
    }

  /* not full, so just add this to the end */
  rgNFS_transitions[it].rg[i].dirid.dev = dirid.dev;
  rgNFS_transitions[it].rg[i].dirid.ino = dirid.ino;
  rgNFS_transitions[it].rg[i].c = 0;
  return i;
}

static int /* index into rgNFS_transitions */
nfs_flt_find(nfs_dirid dirid, int enumHaveToFind)
{ /* add an entry if it's not in there, and inc cNFS_transitions */
  int i = 0;
  for (;i<cNFS_transitions; i++)
    {
    /* FIXGJB: can't test structures for equality?  but can use assignment? */
    if (rgNFS_transitions[i].dirid.dev == dirid.dev &&
	rgNFS_transitions[i].dirid.ino == dirid.ino)
      return i;
    }
  if (enumHaveToFind == NO_CREATE)
    return -1;
  /* fHaveToFind is just there for debugging */
  assert(enumHaveToFind == NEED_NOT_FIND);
  /* not found, so add it */
#ifdef DEBUG_ADDING
  fprintf(stderr, "Adding (%d,%ld)\n",dirid.dev,dirid.ino);
#endif
  rgNFS_transitions[cNFS_transitions].dirid = dirid;
  rgNFS_transitions[cNFS_transitions].c = 0;
  memset(rgNFS_transitions[cNFS_transitions].rg, 0, 
	 NFS_FL_MAX_TOPLEVEL_DIRS * sizeof(nfs_transition_nextdir));
  cNFS_transitions++;
  return i;
}

static void
nfs_async_lookup(kdev_t dev,unsigned long ino, char *name, unsigned long jiffies)
{
  int i;
#ifdef DEBUG_PREDICTIONS
  nfs_fl_dump();
#endif
  /* expire old entries */
  while (ip_hd != ip_tl && 
	 (rgNFS_predictions[ip_hd].jiffies < jiffies - nfs_fl_timeout))
    {
#ifdef DEBUG_PRED_EXPIRE
    fprintf(stderr,"Expriring (%d,%ld) %s @ %ld, now %ld\n",
	    rgNFS_predictions[ip_hd].dirid.dev,
	    rgNFS_predictions[ip_hd].dirid.ino,
	    rgNFS_predictions[ip_hd].name,
	    rgNFS_predictions[ip_hd].jiffies,
	    jiffies);
#endif
    if (rgNFS_predictions[ip_hd].fUsed == FALSE)
      cPredUnused++;
    ip_hd = QUEUE_INC(ip_hd,NFS_PRED_WINDOW_SIZE);
    }

  i = ip_hd;
  while (i != ip_tl)
    {
    if (rgNFS_predictions[i].dirid.dev == dev &&
	rgNFS_predictions[i].dirid.ino == ino &&
	(strncmp(rgNFS_predictions[i].name,name,NFS_FL_MAX_NAME_LEN) == 0))
      {
      /* already on the list, so we don't need to do anything, perhaps
	 if it's about to time out we could refresh it*/
#ifdef DEBUG_PRED_REPEAT
      fprintf(stderr,"Repeat prediction\n");
#endif
      return;
      }
    i = QUEUE_INC(i,NFS_PRED_WINDOW_SIZE);
    }

  if (QUEUE_FULL(ip_hd,ip_tl,NFS_PRED_WINDOW_SIZE))
    {
    cPredOverflows++;
    return;
    }


#ifdef DEBUG_SHOW_PREDICTIONS
  fprintf(stderr,"PREDICT: (%d,%ld) %s @ %ld\n",
	  dev,ino,name,jiffies);
#endif
  cPredMadeTotal++;
  rgNFS_predictions[ip_tl].dirid.dev = dev;
  rgNFS_predictions[ip_tl].dirid.ino = ino;
  rgNFS_predictions[ip_tl].jiffies = jiffies;
  rgNFS_predictions[ip_tl].fUsed = FALSE;
  strncpy(rgNFS_predictions[ip_tl].name,name,NFS_FL_MAX_NAME_LEN);
  ip_tl = QUEUE_INC(ip_tl,NFS_PRED_WINDOW_SIZE);
#ifdef DEBUG_PREDS_LIST
  nfs_fl_dump_preds();
#endif
}

int /* bool */
nfs_predicted_right(kdev_t dev, unsigned long ino, char *name, unsigned long *pjiffies)
{
  /* this searches forwards, should it go backwards? probably
     does not matter, as the list is short, anyway */
  int i = ip_hd;
  while (i != ip_tl)
    {
    if (rgNFS_predictions[i].dirid.dev == dev &&
	rgNFS_predictions[i].dirid.ino == ino &&
	(strncmp(rgNFS_predictions[i].name,name,NFS_FL_MAX_NAME_LEN) == 0) &&
	rgNFS_predictions[i].jiffies + nfs_fl_timeout >= *pjiffies)
      {
      if (rgNFS_predictions[i].fUsed == FALSE)
	{
	cPredCorrect++;
	rgNFS_predictions[i].fUsed = TRUE;
	}
      *pjiffies = rgNFS_predictions[i].jiffies;
      return TRUE;
      }
    i = QUEUE_INC(i,NFS_PRED_WINDOW_SIZE);
    }
  return FALSE;
}

int /* bool */
nfs_still_in_window(kdev_t dev, unsigned long ino, char *name, unsigned long *pjiffies)
{
  int i = QUEUE_DEC(iflc_tl,NFS_FL_WINDOW_SIZE);
  int c = 0;
  /* search backwards */
  do 
    {
    c++;
    if (rgNFS_failures[i].jiffies + nfs_fl_timeout < *pjiffies)
      break;
    if (rgNFS_failures[i].dirid.dev == dev &&
	rgNFS_failures[i].dirid.ino == ino &&
	(strncmp(rgNFS_failures[i].name,name,NFS_FL_MAX_NAME_LEN) == 0))
      {
      *pjiffies = rgNFS_failures[i].jiffies;
#ifdef DEBUG_RECENT_HITS
      fprintf(stderr,"Reuse after %d checks\n",c);
#endif
      return TRUE;
      }
    i = QUEUE_DEC(i,NFS_FL_WINDOW_SIZE);
    }
  while (i != iflc_hd);
#ifdef DEBUG_RECENT_HITS
  fprintf(stderr,"Failure after %d checks\n",c);
#endif
  return FALSE;
}


int /* number of predictions made */
nfs_fl_make_predictions(kdev_t dev, unsigned long ino, char *name, unsigned long jiffies)
{ /* don't make predictions that we've already made */
  /* also pick some threshold only above-which we'll do the rpc call for the prediction */
  /* make no more than NFS_FL_MAX_PREDICTIONS per call */
  int it;
  int cPredMade = 0;
  int itarg = 0;
  int tot;
  nfs_dirid di;
  di.dev = dev; di.ino = ino;
  it = nfs_flt_find(di,NO_CREATE);
  if (it < 0)
    return;
  if ((tot = rgNFS_transitions[it].cTotTransitions) == 0)
    return;
  while (cPredMade < NFS_FL_MAX_PREDICTIONS)
    {
    int p = 100 * rgNFS_transitions[it].rg[itarg].c / tot;
    if (p > NFS_P_REQUIRED_TO_PREDICT) 
      {
      nfs_async_lookup(rgNFS_transitions[it].rg[itarg].dirid.dev,
		       rgNFS_transitions[it].rg[itarg].dirid.ino,name,jiffies);
      cPredMade++;
      } 
    else 
      break;
    itarg++;
    }
}

static void
nfs_transitions_inc_and_order(int it,nfs_dirid diTarg)
{
  int iTarg;
  rgNFS_transitions[it].cTotTransitions++;
#ifdef DEBUG_INC_DEC
  fprintf(stderr,"Inc spot %d, (%d,%ld), to (%d,%ld)\n",
	  it,rgNFS_transitions[it].dirid.dev,rgNFS_transitions[it].dirid.ino,
	  diTarg.dev,diTarg.ino);
#endif
  iTarg = nfs_flt_target_find(it, diTarg, NEED_NOT_FIND);
  rgNFS_transitions[it].rg[iTarg].c++;
  if (iTarg > 0 && 
      rgNFS_transitions[it].rg[iTarg].c > rgNFS_transitions[it].rg[iTarg-1].c)
    {
    nfs_transition_nextdir temp = rgNFS_transitions[it].rg[iTarg];
    int i = iTarg - 1;
    while (i > 0 &&
	   rgNFS_transitions[it].rg[iTarg].c > rgNFS_transitions[it].rg[i-1].c)
      i--;
    /* now swap */
    rgNFS_transitions[it].rg[iTarg] = rgNFS_transitions[it].rg[i];
    rgNFS_transitions[it].rg[i] = temp;
    }
}

static void
nfs_transitions_dec_and_order(int it,nfs_dirid diTarg)
{
  int iTarg;
  rgNFS_transitions[it].cTotTransitions--;
#ifdef DEBUG_INC_DEC
  fprintf(stderr,"Dec spot %d, (%d,%ld), to (%d,%ld)\n",
	  it,rgNFS_transitions[it].dirid.dev,rgNFS_transitions[it].dirid.ino,
	  diTarg.dev,diTarg.ino);
#endif
  iTarg = nfs_flt_target_find(it, diTarg, MUST_FIND);
  if (--rgNFS_transitions[it].rg[iTarg].c == 0) 
    {
    rgNFS_transitions[it].rg[iTarg].dirid.dev = 0;
    rgNFS_transitions[it].rg[iTarg].dirid.ino = 0;
#ifdef DEBUG_RESET_TARGET
    fprintf(stderr,"Reset target offset %d\n",iTarg);
#endif
    }

  while (iTarg < NFS_FL_MAX_TRANSITION_DIRS &&
	 rgNFS_transitions[it].rg[iTarg].c < rgNFS_transitions[it].rg[iTarg+1].c)
    {
    /* swap them */
    nfs_transition_nextdir temp = rgNFS_transitions[it].rg[iTarg];
#ifdef DEBUG_MOVING
    fprintf(stderr,"Moving %d downto %d\n",iTarg+1,iTarg);
#endif
    rgNFS_transitions[it].rg[iTarg] = rgNFS_transitions[it].rg[iTarg+1];
    rgNFS_transitions[it].rg[iTarg+1] = temp;
    iTarg++;
#ifdef DEBUG_MOVING
    fprintf(stderr,"Now %d vs %d\n",
	    rgNFS_transitions[it].rg[iTarg].c,
	    rgNFS_transitions[it].rg[iTarg+1].c);
#endif
    }
}

/* we pass in the index, since the fn needs to access the last several
   entries as well */
static void
nfs_fl_update(int iflc)
{
  int it = 0;
  /* update data structures for adding rgNFS_failures[iflc] */
#ifdef DEBUG_QUEUE
  fprintf(stderr,"Updating %d\n",iflc);
#endif
  it = nfs_flt_find(rgNFS_failures[iflc].dirid,NEED_NOT_FIND);
  rgNFS_transitions[it].c++;
  /* see if we need to bubble it up */
  if (it > 0 && rgNFS_transitions[it].c > rgNFS_transitions[it-1].c)
    {
    nfs_transitions temp = rgNFS_transitions[it];
    int i = it - 1;
    while (i > 0 && rgNFS_transitions[it].c > rgNFS_transitions[i-1].c) i--;
    /* now swap them (not a stable ordering, but more efficient) */
    rgNFS_transitions[it] = rgNFS_transitions[i];
    rgNFS_transitions[i] = temp;
    it = i;
    }

  /* now we need to look through the rgNFS_failures before iflc (wrapping circularly)
     and increment the transition information to diAdd, this new directory */
  { /* scope */
  int iflcTarg = iflc;
  char *nameTarg = rgNFS_failures[iflcTarg].name;
  nfs_dirid diTarg = rgNFS_failures[iflcTarg].dirid;
  int i = 0;
  for ( ; i < NFS_FL_LOOKAHEAD; i++ )
    {
    iflc = QUEUE_DEC(iflc,NFS_FL_WINDOW_SIZE);
    /* if it's a lookup of the same name in a different directory, we need to increment */
    if ((strncmp(rgNFS_failures[iflc].name,nameTarg,NFS_FL_MAX_NAME_LEN) == 0) &&
	(rgNFS_failures[iflc].dirid.dev != diTarg.dev ||
	 rgNFS_failures[iflc].dirid.ino != diTarg.ino) &&
	rgNFS_failures[iflc].jiffies + nfs_fl_timeout >= rgNFS_failures[iflcTarg].jiffies)
      {
      int it = nfs_flt_find(rgNFS_failures[iflc].dirid,MUST_FIND);
      nfs_transitions_inc_and_order(it,diTarg);
      }
#ifdef DEBUG_UPDATE_FAILS
    else
      {
#define bad(x) fputs(x,stderr)
      bad("| ");
      (strncmp(rgNFS_failures[iflc].name,nameTarg,NFS_FL_MAX_NAME_LEN) == 0) || bad("bad match ");
      (rgNFS_failures[iflc].dirid.dev != diTarg.dev ||
       rgNFS_failures[iflc].dirid.ino != diTarg.ino) || bad("same dir ");
      (rgNFS_failures[iflc].jiffies + nfs_fl_timeout >= rgNFS_failures[iflcTarg].jiffies)
	|| bad("timeout ");
      bad(" |\n");
#undef bad
      }
#endif
    }
  } /* end scope */
}

/* we pass in the index, since the fn needs to access the next several
   entries as well */
static void
nfs_fl_remove(int iflc)
{
  int it;
  /* update data structures for removing rgNFS_failures[iflc] */
#ifdef DEBUG_QUEUE
  fprintf(stderr,"Removing %d\n",iflc);
#endif
  it = nfs_flt_find(rgNFS_failures[iflc].dirid,MUST_FIND);
  if (--rgNFS_transitions[it].c == 0)
    {
    cNFS_transitions--;
    /* and move the one from the last spot to here */
    rgNFS_transitions[it] = rgNFS_transitions[cNFS_transitions];
    rgNFS_transitions[cNFS_transitions].dirid.dev = 0;
    rgNFS_transitions[cNFS_transitions].dirid.ino = 0;
    memset(rgNFS_transitions[cNFS_transitions].rg, 0, 
	   NFS_FL_MAX_TOPLEVEL_DIRS * sizeof(nfs_transition_nextdir));
    return;
    }

  /* it has a non-zero count still */
  if (rgNFS_transitions[it].c < rgNFS_transitions[it+1].c)
    { /* bubble it down */
    nfs_transitions temp = rgNFS_transitions[it];
    int i = it+1;
    while (i+1 < cNFS_transitions &&
	   rgNFS_transitions[it].c < rgNFS_transitions[i+1].c) i++;
    /* now swap them (not a stable ordering, but more efficient) */
    rgNFS_transitions[it] = rgNFS_transitions[i];
    rgNFS_transitions[i] = temp;
    it = i;
    }

  /* now we need to look through the rgNFS_failures after iflc (wrapping circularly)
     and decrement the transition information from rgNFS_transitions[iflc].rg */
  { /* scope */
  int iflcSrc = iflc;
  char *name = rgNFS_failures[iflcSrc].name;
  int i = 0;
  for ( ; i < NFS_FL_LOOKAHEAD; i++)
    {
    iflc = QUEUE_INC(iflc,NFS_FL_WINDOW_SIZE);
    /* if it's a lookup of the same name in a different directory, we need to decrement */
    if (strncmp(rgNFS_failures[iflc].name,name,NFS_FL_MAX_NAME_LEN) == 0 &&
	(rgNFS_failures[iflc].dirid.dev != rgNFS_failures[iflcSrc].dirid.dev ||
	 rgNFS_failures[iflc].dirid.ino != rgNFS_failures[iflcSrc].dirid.ino) &&
	rgNFS_failures[iflc].jiffies <= rgNFS_failures[iflcSrc].jiffies + nfs_fl_timeout)
      {
      nfs_transitions_dec_and_order(it,rgNFS_failures[iflc].dirid);
      }
    }
  } /* end scope */
}

static void
nfs_fl_add(kdev_t dev, unsigned long ino, char *name, unsigned long jiffies)
{
  /* I could choose to not add things already in the cache,
     but that would take a linear scan of the list and doesn't seem
     to be worth the trouble */
  if (strlen(name) > NFS_FL_MAX_NAME_LEN)
    return;
  if (iflc_hd == QUEUE_INC(iflc_tl,NFS_FL_WINDOW_SIZE))
    { /* already full, so remove first */ 
    nfs_fl_remove(iflc_hd);
    iflc_hd = QUEUE_INC(iflc_hd,NFS_FL_WINDOW_SIZE);
    }
  rgNFS_failures[iflc_tl].dirid.dev = dev;
  rgNFS_failures[iflc_tl].dirid.ino = ino;
  strncpy(rgNFS_failures[iflc_tl].name,name,NFS_FL_MAX_NAME_LEN);
  rgNFS_failures[iflc_tl].jiffies = jiffies;
  /* now update data structure for adding rgNFS_failures[iflc_tl] */
  nfs_fl_update(iflc_tl);
  iflc_tl = QUEUE_INC(iflc_tl,NFS_FL_WINDOW_SIZE);
}

static void
nfs_fl_dump_preds()
{
  int i = ip_hd;
  fprintf(stderr,"-----\nPredictions (%d)\n", 
	  QUEUE_SIZE(ip_hd,ip_tl,NFS_PRED_WINDOW_SIZE));
  while (i != ip_tl)
    {
    fprintf(stderr,"(%d,%ld) %s @ %ld\n",
	    rgNFS_predictions[i].dirid.dev,
	    rgNFS_predictions[i].dirid.ino,
	    rgNFS_predictions[i].name,
	    rgNFS_predictions[i].jiffies);
    i = QUEUE_INC(i,NFS_PRED_WINDOW_SIZE);
    }
}



static void
nfs_fl_dump()
{
  int i = iflc_hd;
  fprintf(stderr,"Queue window of failed lookups (oldest first)\n");
  while (i != iflc_tl)
    {
    fprintf(stderr,"(%d,%ld) %s @ %ld\n",
	    rgNFS_failures[i].dirid.dev,
	    rgNFS_failures[i].dirid.ino,
	    rgNFS_failures[i].name,
	    rgNFS_failures[i].jiffies);
    i = QUEUE_INC(i,NFS_FL_WINDOW_SIZE);
    }

  fprintf(stderr,"\nDirectories in recent list\n");
  for (i = 0; i < cNFS_transitions; i++)
    {
    int id = 0;
    fprintf(stderr,"(%d,%ld) has count %d with %d transitions\n",
	    rgNFS_transitions[i].dirid.dev,
	    rgNFS_transitions[i].dirid.ino,
	    rgNFS_transitions[i].c,
	    rgNFS_transitions[i].cTotTransitions);
    while (id < NFS_FL_MAX_TRANSITION_DIRS)
      {
      nfs_transition_nextdir *nfs_nd = &rgNFS_transitions[i].rg[id++];
      if (nfs_nd->dirid.dev == 0)
	break;
      fprintf(stderr,"  -> (%d,%ld), %d\n",
	      nfs_nd->dirid.dev,
	      nfs_nd->dirid.ino,
	      nfs_nd->c);
      }
    }
}

/* Testing infrastructure below */

static double start_time = 0.0;

unsigned long JiffiesFromTime(double time)
{
  return (time-start_time)*HZ;
}

int main(int argc, char *argv[])
{
  char szLine[512];
  int dev;
  unsigned long ino;
  char szName[256];
  unsigned long sec;
  int cs;
  double time;
  unsigned long jiffies;
  unsigned long newjiffies;
  int fAlreadyHaveIt = FALSE;
  fprintf(stderr,"Using %d+%d bytes\n",sizeof(rgNFS_transitions),sizeof(rgNFS_failures));
  

  while(fgets(szLine,512,stdin) &&
	sscanf(szLine,"(%d/%lu, %s %d.%d, type",&dev,&ino,szName,&sec,&cs) > 0) 
    {
    szName[strlen(szName)-1] = 0;
    time = sec + (double) cs/100;
#ifdef DEBUG_PARSE
    fprintf(stderr,"Got %d %ld %s %f\n",dev,ino,szName,time);
#endif
    if (start_time = 0.0)
      start_time = time;

    jiffies = JiffiesFromTime(time);
    newjiffies = jiffies;
    if (nfs_predicted_right(dev,ino,szName,&newjiffies))
      {
      cPredHits++;
      fAlreadyHaveIt = TRUE;
      }
    else 
      {
      /* Only check if it's in the window if it wasn't a prediction */
      if (nfs_still_in_window(dev,ino,szName,&newjiffies))
	{
	cRecentUseHit++;
	fAlreadyHaveIt = TRUE;
	}
      }
    cTotLookups++;
    nfs_fl_make_predictions(dev,ino,szName,jiffies);
    nfs_fl_add(dev,ino,szName,jiffies);
#ifdef DEBUG_LISTS
    nfs_fl_dump();
    fprintf(stderr,"\n-----\n");
#endif
#ifdef DEBUG_PREDS_LIST
    nfs_fl_dump_preds();
#endif
    }

#ifdef DEBUG_LISTS
  nfs_fl_dump();
#endif
#ifdef DEBUG_PREDS_LIST
  nfs_fl_dump_preds();
#endif
  fprintf(stderr,"\nMade %d, Correct %d, Unused %d, Hits %d, RecentUseHit %d, TotalLookups %d\n",
	  cPredMadeTotal, cPredCorrect, cPredUnused, cPredHits, cRecentUseHit,cTotLookups);
  fprintf(stderr,"PredOverflows %d\n",cPredOverflows);
  fprintf(stderr,"Pct Used = %2.2f; Pct Hits = %2.2f\n",
	  (double) cPredCorrect*100.0/(cPredCorrect+cPredUnused),
	  (double) (cRecentUseHit+cPredHits)*100.0/cTotLookups);
  exit(0);
}
