/*
 *  Disk Caching NFS: CSE 551 Spring 1997  -  University of Washington
 *
 * $Id: nccd.c,v 1.5 1997/06/11 07:50:56 acollins Exp $
 */

char usageInfo[] = "\
USAGE:
   nccd [-h]               display this help message\n\
        [-d <root dir>]    specify cache directory   (default: /.nfs-cache)\n\
        [-w <watch file>]  file to watch (full path) (default: <root>/.FULL)\n\
        [-l <log file>]    log file (full path)      (default: no logging)\n\
";

/*    General info:
 *
 * if you #define DEBUG_STUB_DEV, then it will write to /tmp instead
 * of /dev (for testing on !gumshoe).
 *
 * this version currently considers only files from the filesystem it has
 * been instructed to clean.  It sorts by atime and removes as many files
 * as needed to get to 80 (512 byte) blocks.  If blocks need to be 4196 or
 * 8192, changing BLOCK_SIZE should do it.
 *
 * **** this version has lots of screen output ****
 */

#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>

#define MAX_FN_LEN 256
#define BLOCK_SIZE 4096

#ifdef DEBUG_STUB_DEV
#define debug_printf(x...) printf(x)
#else
#define debug_printf(x...) printf(x)
#endif

static FILE* fpLog = NULL;
#define log_printf(x...) ((fpLog != NULL) ? fprintf (fpLog, x) : 0)


/* Define a linked-list of files */
typedef struct _FileRec
{
   char fn[MAX_FN_LEN];
   time_t atime;
   time_t mtime;
   long blocks;
   mode_t mode;

   int mark;

   char host[MAX_FN_LEN];
   int device;
   int inode;

   struct _FileRec* next;
} FileRec;

FileRec* GetCacheFiles (char* cachePath, char* hostToClean, int deviceToClean,
                        int* numFiles)
{
   struct stat statbuf;
   struct dirent* dirp;
   DIR* dir;
   int ret;
   char* ptr;

   FileRec* retval;
   retval = NULL;
   *numFiles = 0;

   debug_printf ("GetCacheFiles\n");

   if (cachePath[strlen (cachePath)-1] == '/')
      cachePath[strlen (cachePath)-1] = 0;
   if (lstat (cachePath, &statbuf) < 0)
      {
      exit (1);
      }
   if (S_ISDIR (statbuf.st_mode) == 0)
      {
      exit (1);
      }

   strcat (cachePath, "/");
   dir = opendir (cachePath);
   if (!dir)
      {
      exit (1);
      }
   while ((dirp = readdir (dir)) != NULL)
      {
      if (dirp->d_name[0] != '.')
         {
         struct stat sb;
         char tempFn[MAX_FN_LEN];

         strcpy (tempFn, cachePath);
         strcat (tempFn, dirp->d_name);

         debug_printf ("Looking at: '%s'\n", tempFn);

         lstat (tempFn, &sb);

         if ((!S_ISDIR (sb.st_mode)) && (sb.st_mode & S_IEXEC))
            {
            FileRec* newFile = (FileRec*)malloc (sizeof (FileRec));

            strcpy (newFile->fn, dirp->d_name);
            newFile->atime = sb.st_atime;
            newFile->mtime = sb.st_mtime;

            /* use data blocks only */
            /* newFile->blocks = sb.st_blocks; */
            newFile->blocks = (sb.st_size / BLOCK_SIZE) + 
                              ((sb.st_size % BLOCK_SIZE) ? 1 : 0);

            newFile->mode = sb.st_mode;
            strcpy (newFile->host, strtok (newFile->fn, ","));
            newFile->device = atoi (strtok (NULL, "-"));
            newFile->inode = atoi (strtok (NULL, "-"));
           
            strcpy (newFile->fn, dirp->d_name);
            if ((hostToClean[0] == 0) || 
                (!strcmp (newFile->host, hostToClean) && 
                 newFile->device == deviceToClean))
               {
               newFile->next = retval;
               retval = newFile;
               debug_printf ("Considering file: '%s' (%d)\n", newFile->fn, 
                       newFile->atime);
               ++(*numFiles);
               }
            else
               {
               debug_printf ("Rejecting file: '%s'\n", newFile->fn);
               free (newFile);
               }
            }
         }
      }

   closedir (dir);
   return retval;
}

typedef struct _CorrectionRec
{
   char host[MAX_FN_LEN];
   int device;
   int delta;

   struct _CorrectionRec* next;
} CorrectionRec;

CorrectionRec* BuildCorrectionList (FileRec* files)
{
   FileRec* temp;
   CorrectionRec* corrections;

   corrections = NULL;
   temp = files;
   while (temp)
      {
      CorrectionRec* t2;

      /* update or add to corrections */
      t2 = corrections;
      while (t2)
         {
         if (!strcmp (t2->host, temp->host) &&
             t2->device == temp->device)
            {
            t2->delta -= temp->blocks;
            break;
            }
         t2 = t2->next;
         }
      if (!t2)
         {
         t2 = (CorrectionRec*) malloc (sizeof (CorrectionRec));
         strcpy (t2->host, temp->host);
         t2->device = temp->device;
         t2->delta = -temp->blocks;
         t2->next = corrections;
         corrections = t2;
         }

      temp = temp->next;
      }

   return corrections;
}

void VictimizeCacheFiles (FileRec* files, char* cachePath)
{
   int blocksFreed;
   int fd;
   char buf[MAX_FN_LEN];
   CorrectionRec* corrections;

   corrections = BuildCorrectionList (files);
   debug_printf ("VictimizeCacheFiles\n");

   while (files)
      {
      FileRec* temp;

      char tempFn[MAX_FN_LEN];
      strcpy (tempFn, cachePath);
      strcat (tempFn, files->fn);
      files->mode &= ~S_IEXEC;
      chmod (tempFn, files->mode);
      unlink (tempFn);

      log_printf ("   unlinked file: '%s' (%d pages)\n", 
                  files->fn, files->blocks);

      temp = files;
      files = files->next;
      free (temp);
      }

   /* tell the server */
#ifndef DEBUG_STUB_DEV
   fd = open ("/dev/nfs-cache-space", O_WRONLY | O_APPEND);
#else
   fd = open ("/tmp/nfs-cache-space", O_WRONLY | O_APPEND);
#endif
   while (corrections)
      {
      CorrectionRec* t2;

      t2 = corrections;
      sprintf (buf, "%s,%d,-%d\n", 
               corrections->host, 
               corrections->device, 
               -corrections->delta);
      write (fd, buf, strlen (buf));

      log_printf ("   wrote to /dev/nfs-cache-space: '%s'\n", buf);

      corrections = corrections->next;
      free (t2);
      }
   close (fd);
}

int CompareFunc (void* p1, void* p2)
{
   FileRec* f1;
   FileRec* f2;
   int retval;

   f1 = *(FileRec**)p1;
   f2 = *(FileRec**)p2;
   
   if (f1->atime < f2->atime)
      retval = -1;
   else if (f1->atime > f2->atime)
      retval = 1;
   else
      retval = 0;

   debug_printf ("CompareFunc: %d, %d -> %d\n", f1->atime, f2->atime, retval);
   return retval;
}

FileRec* ChooseVictimsAndFreeOthers (FileRec* files, int numFiles, 
                                     int blocksToMark)
{
   FileRec** prev;
   FileRec*  curs;
   int i;
   FileRec** fileArray;
   int markedBlocks;

   debug_printf ("ChooseVictimsAndFreeOthers %d files\n", numFiles);

   fileArray = (FileRec**) malloc (numFiles * sizeof (FileRec*));
   i = 0;
   curs = files;
   while (curs)
      {
      fileArray[i] = curs;      

      ++i;
      curs = curs->next;
      }
   qsort (fileArray, numFiles, sizeof (FileRec*), CompareFunc);

   i = 0;
   markedBlocks = 0;
   while (i < numFiles && markedBlocks < blocksToMark)
      {
      debug_printf ("Marking: '%s'\n", fileArray[i]->fn);
      fileArray[i]->mark = 1;
      markedBlocks += fileArray[i]->blocks;
      ++i;
      }  
   while (i < numFiles)
      {
      debug_printf ("Not marking: '%s'\n", fileArray[i]->fn);
      fileArray[i]->mark = 0;
      ++i;
      }

   prev = &files;
   curs = files;
   while (curs)
      {
      if (curs->mark == 0)
         {
         FileRec* temp;
         debug_printf ("sparing: '%s'\n", curs->fn);
         temp = curs;
         curs = curs->next;
         *prev = curs;
         free (temp);
         }
      else
         {
         debug_printf ("killing: '%s'\n", curs->fn);
         prev = &curs->next;
         curs = curs->next;
         }
      }

   free (fileArray);
   return files;
}

void DoCleaning (char* cacheDir, int timeStamp, char* hostToClean, 
                 int deviceToClean)
{
   int numFiles;
   int blocksToClean;
   FileRec* list;
   FILE* fpMounts;
   char buf[512];
   int totalSpace;
   int usedSpace;

   totalSpace = 0;
   usedSpace = 0;

 #ifndef DEBUG_STUB_DEV 
   fpMounts = fopen ("/proc/net/nfs-mounts", "r");
#else
   fpMounts = fopen ("/tmp/nfs-mounts", "r");
#endif
   if (!fpMounts)
      {
      exit (1);
      }
   /* skip the first two lines */
   fgets (buf, 512, fpMounts);
   fgets (buf, 512, fpMounts);
   while (fgets (buf, 512, fpMounts))
      {
      if (buf[0] == '#')
         {
         }
      else
         {
         char host[MAX_FN_LEN];
         char* chp;
         int device;

	 printf ("parsing line: %s", buf);
         strcpy (host, strtok (buf, ","));
         chp = strtok (NULL, ":");
         if (chp)
            device = atoi (chp);
         else
            device = -1;

         if ((hostToClean[0] == 0) || 
             (!strcmp (hostToClean, host) && device == deviceToClean))
            {
            int used;
            int total;

            if (!strtok (NULL, " \n"))   /* has */
               continue;

            chp = strtok (NULL, "/\n");
            used = chp ? atoi (chp) : 0;
            chp = strtok (NULL, " \n");
            total = chp ? atoi (chp) : 0;

            strtok (NULL, " ");   /* pages */
            strtok (NULL, " ");   /* used  */
            strtok (NULL, " ");   /* on    */
            chp = strtok (NULL, " \n");
            if (chp && chp[strlen (chp)-1] == '/')
               chp[strlen (chp)-1] = 0;

            if (chp && !strncmp (chp, cacheDir, strlen (cacheDir)-1))
               {
               usedSpace += used;
               totalSpace += total;
               debug_printf ("incorporated %s,%d in '%s'  %d/%d\n",
                       host, device, cacheDir, used, total);
               } 
            else
               debug_printf ("did not incorporate %s,%d in '%s' != cacheDir '%s'  %d/%d\n",
                       host, device, chp, cacheDir, used, total);
            }
         }
      }
   fclose (fpMounts);

   blocksToClean = (int)(0.1 * (double) usedSpace);

   debug_printf ("===> %d/%d total, cleaning %d\n", usedSpace, totalSpace, 
           blocksToClean);

   /* Choose and delete victim files */
   list = GetCacheFiles (cacheDir, hostToClean, deviceToClean, &numFiles),
   list = ChooseVictimsAndFreeOthers (list, numFiles, blocksToClean);
   VictimizeCacheFiles (list, cacheDir);

   debug_printf ("Done, numFiles is now: %d\n", numFiles);

}

void DoStartupStuff (char* cacheDir)
{
   int fd;
   char buf[MAX_FN_LEN];
   CorrectionRec* corrections;
   int numFiles;
   FileRec* list;

   debug_printf ("DoStartupStuff\n");
   log_printf ("startup, scanning cache...\n");

   list = GetCacheFiles (cacheDir, "", 0, &numFiles);
   corrections = BuildCorrectionList (list);

   /* tell the server */
#ifndef DEBUG_STUB_DEV
   fd = open ("/dev/nfs-cache-space", O_WRONLY | O_APPEND);
#else
   fd = open ("/tmp/nfs-cache-space", O_WRONLY | O_APPEND);
#endif

   while (corrections)
      {
      CorrectionRec* t2;

      t2 = corrections;
      sprintf (buf, "%s,%d,+%d\n",              /* may become = */
               corrections->host, 
               corrections->device, 
               -corrections->delta);
      write (fd, buf, strlen (buf));

      log_printf ("   wrote to /dev/nfs-cache-space: '%s'\n", buf);

      corrections = corrections->next;
      free (t2);
      }
   close (fd);

   /* free the file list */
   while (list)
      {
      FileRec* temp = list;
      list = list->next;
      free (temp);
      }
}

int main(int argc, char* argv[])
{
   char cacheDir[MAX_FN_LEN];
   char watchFile[MAX_FN_LEN];
   char hostToClean[MAX_FN_LEN];
   int deviceToClean;
   char prefix[128];
   FILE* fp;
   int deviceId = 0;
   int timeStamp = 0;
   char buf[256];
   int bytes;
   struct stat sb;
   int oldSize;

   strcpy (cacheDir, "/.nfs-cache");
   strcpy (watchFile, "");
   deviceToClean = 0;
   strcpy (prefix, "");

   for (;;)
      {
      int c;
      c = getopt (argc, argv, "d:w:l:h");
      if (c == -1)
         break;

      switch (c)
         {
         case 'h':
            fprintf (stderr, "NFS Cache Cleaning Daemon (nccd):\n\n");
            fprintf (stderr, "%s\n", usageInfo);
            exit (0);
            break;

         case 'd':
            strcpy (cacheDir, optarg);
            if (cacheDir[strlen (cacheDir)-1] != '/')
               {
               cacheDir[strlen (cacheDir)+1] = 0;
               cacheDir[strlen (cacheDir)]   = '/';
               }
            break;

         case 'w':
            strcpy (watchFile, optarg);
            break;

         case 'l':
            fpLog = fopen (optarg, "a");
            if (!fpLog)
               {
               fprintf (stderr, "ERROR: can't open log file: '%s'\n", optarg);
               exit (1);
               }
            break;
         }
      }

   if (watchFile[0] == 0)
      sprintf (watchFile, "%s.FULL", cacheDir);

   debug_printf ("cacheDir: '%s'\ndeviceId: %d\nwatchFile: '%s'\nprefix: '%s'\n",
           cacheDir, deviceId, watchFile, prefix);

   DoStartupStuff (cacheDir);

   fp = fopen (watchFile, "r");
   while (fgets (buf, 255, fp)) ;       /* skip old stuff */

   if (lstat (watchFile, &sb) < 0)
      {
      exit (1);
      }
   sb.st_mode &= ~S_IXUSR;
   sb.st_mode &= ~S_IXGRP;
   chmod (watchFile, sb.st_mode);

   oldSize = sb.st_size;
   
   /* Wait for the kernel to wake us up */
   for (;;)
      {
      sleep (1);

      if (lstat (watchFile, &sb) < 0)
         {
         exit (1);
         }

      if (sb.st_size > oldSize)
         {
         while (fgets (buf, 255, fp))
            {
            char* chp;

            log_printf ("got message: '%s'\n", buf);

            timeStamp = atoi (strtok (buf, " "));
            chp = strtok (NULL, ",");
            if (chp)
               {
               strcpy (hostToClean, chp);
               deviceToClean = atoi (strtok (NULL, " "));
 
               debug_printf ("max size violation on %s,%d\n", 
                       hostToClean, deviceToClean);
               DoCleaning (cacheDir, timeStamp, hostToClean, 
                           deviceToClean);
               }
            else if (timeStamp > 0)
               {
               debug_printf ("min free violation\n");
               DoCleaning (cacheDir, timeStamp, "", 0);
               }
            else
               debug_printf ("BAD\n");
            }

         /* Re-enable caching */
         sb.st_mode &= ~S_IXGRP;
         sb.st_mode &= ~S_IXUSR;
         chmod (watchFile, sb.st_mode);
         }
      }
   fclose (fp);
   if (fpLog)
      fclose (fpLog);
}
